;;
;; scmail - a mail filter written in Scheme
;;
;; Copyright (C) 2002 Satoru Takabayashi <satoru@namazu.org> 
;;     All rights reserved.
;;     This is free software with ABSOLUTELY NO WARRANTY.
;;
;; You can redistribute it and/or modify it under the terms of 
;; the GNU General Public License version 2.
;;

(define-module scmail
  (use srfi-1)
  (use srfi-13)
  (use gauche.charconv)
  (use gauche.regexp)
  (use gauche.parseopt)
  (use rfc.base64)
  (use rfc.quoted-printable)
  (use gauche.net)
  (use file.util)
  (export <config> <mail>
	  expand-file-name file-list file-name
	  mail-filter mail-removed? mail-copy mail-query mail-remove
	  command-copy command-refile command-forward
	  read-config-file read-kill-file add-filter-rules add-filter-rule! 
	  set-match-data-replace-rule! 
          scmail-command-log scmail-main
	  ))

(select-module scmail)


;; For avoiding errors while handling incomplete string
(define (safe-rxmatch pattern string)
    (with-error-handler
     (lambda (e) #f)
     (lambda ()
       (rxmatch pattern string))))

;;
;; FIXME: Reinvent it because filter in srfi-1 module is too slow.
;;
(define (filter predicate items)
  (let loop ((result '()) 
	     (items items))
    (cond ((null? items)
	   (reverse! result))
	  ((predicate (car items))
	   (loop (cons (car items) result) (cdr items)))
	  (else (loop result (cdr items))))))


;; (expand-file-name "~/Mail") => "/home/foobar/Mail"
;; (expand-file-name "bar" "/home/foo") => "/home/foo/bar"
(define (expand-file-name name . directory)
  (if (null? directory)
      (if (eq? (string-ref name 0) #\~)
	  (string-append (sys-getenv "HOME") (string-drop name 1))
	  name)
      (string-append (car directory) "/" name)))

;; dispatch function
(define (file-list config directory)
  ((slot-ref config 'file-list) directory))

;; dispatch function
(define (file-name config id directory)
  ((slot-ref config 'file-name) id directory))

;; dispatch function
(define (mail-copy config mail mailbox folder)
  ((slot-ref config 'mail-copy) mail mailbox folder))

;; (mh-file-list "~/Mail/inbox") => (1 2 3 4 5 6 7 8)
(define (mh-file-list directory)
  (sort
   (map string->number
	(filter (lambda (x) (safe-rxmatch #/^[0-9]+$/ x))
		(directory-list directory)))))

;; (mh-file-name 123 "~/Mail/inbox") => "/home/foobar/Mail/inbox/123"
(define (mh-file-name id directory)
  (expand-file-name
   (number->string id)
   (expand-file-name directory)))

(define (mh-mail-copy mail mailbox folder)
  (let* ((dest-directory (make-dest-directory mailbox folder))
	 (new-id (+ 1 (mh-last-mail-id dest-directory)))
	 (file-name (mh-file-name new-id dest-directory)))
    (call-with-output-file file-name
      (lambda (port)
	(mail-write mail port)))
    file-name))

;; (mh-last-mail-id "~/Mail/inbox") => 8
(define (mh-last-mail-id directory)
  (let ((mlist (mh-file-list directory)))
    (if (null? mlist)
	0
	(car (last-pair mlist)))))

(define (maildir-file-list directory)
  (maildir-make-sub-directories directory)
  (let ((directory (expand-file-name "cur" directory)))
    (sort
     (filter (lambda (x) 
               (and (lambda (x) (safe-rxmatch #/^[0-9]+/ x))
                    (file-is-regular? (expand-file-name x directory))))
             (directory-list directory)))))

(define (maildir-file-name id directory)
  (expand-file-name
   id
   (expand-file-name "cur"
                     (expand-file-name directory))))

(define (maildir-mail-copy mail mailbox folder)
  (let* ((dest-directory (make-dest-directory mailbox folder))
	 (new-id (maildir-generate-new-id))
	 (file-name (maildir-file-name new-id dest-directory)))
    (maildir-make-sub-directories dest-directory)
    (if (equal? folder "") ;; the top of ~/Maildir
        (maildir-safe-mail-copy mail dest-directory new-id)
        (call-with-output-file file-name
          (lambda (port)
            (mail-write mail port) file-name)))))

(define (maildir-safe-mail-copy mail directory id)
  (let ((timer 30)
        (tmp-file-name 
         (expand-file-name id (expand-file-name "tmp" directory)))
        (new-file-name 
         (expand-file-name id (expand-file-name "new" directory))))
    (while (file-exists? tmp-file-name)
           (if (<= timer 0)
               (errorf "maildir-safe-mail-copy timeout!"))
           (sys-sleep 2)
           (set! timer (- timer 2)))
    (call-with-output-file tmp-file-name
      (lambda (port)
        (mail-write mail port)))
    (sys-link tmp-file-name new-file-name)
    (sys-unlink tmp-file-name)
    new-file-name))

(define (maildir-make-sub-directories directory)
  (create-directory* (expand-file-name "tmp" directory))
  (create-directory* (expand-file-name "cur" directory))
  (create-directory* (expand-file-name "new" directory)))

(define maildir-id 0)
(define (maildir-generate-new-id)
  (set! maildir-id (+ maildir-id 1))
  (format "~a.~a_~a.~a"
          (sys-time)
          (sys-getpid)
          maildir-id
          (sys-gethostname)))

;; (decode-field "=?iso-2022-jp?B?GyRCJCIbKEI=?=\n abc" "eucjp") => "abc"
;; (decode-field "=?iso-2022-jp?Q?=1B=24=42=24=22=1B=28=42?=abc" "eucjp")
;;  => "abc"
(define (decode-field str to-code)
  (with-error-handler 
   (lambda (e) str)
   (lambda ()
     (regexp-replace-all #/=\?([^?]+)\?([BQ])\?([^?]+)\?=\s*/ 
			 str
			 (lambda (m)
			   (let* ((charcode (rxmatch-substring m 1))
				  (encoding (rxmatch-substring m 2))
				  (message  (rxmatch-substring m 3))
				  (decode (if (equal? encoding "B")
					      base64-decode-string
					      quoted-printable-decode-string)))
			     (ces-convert (decode message)
					  charcode
					  to-code)))))))

;; (send-mail "localhost" 25 (open-input-string "Hello!")
;;  "from@localhost" "to@localhost")
(define (send-mail host port iport mail-from recipients)
  (with-error-handler
   (lambda (e) (errorf "send-mail failed: ~a" (slot-ref e 'message)))
   (lambda ()
     (call-with-client-socket
      (make-client-socket 'inet host port)
      (lambda (in out)
	(let ((send-command 
	       (lambda (command code)
		 (when command (format out "~a\r\n" command))
		 (let* ((line (read-line in))
			(return-code (string->number (substring line 0 3))))
		   (if (eq? return-code code)
		       line
		       (errorf "smtp-error: ~a => ~a" command line))))))
	  (send-command #f 220)
	  (send-command (format "HELO ~a" (sys-gethostname)) 250)
	  (send-command (format "MAIL FROM: <~a>" mail-from) 250)
	  (for-each (lambda (rcpt)
		      (send-command (format "RCPT TO: <~a>" rcpt) 250))
		    (if (string? recipients) (list recipients) recipients))
	  (send-command "DATA" 354)
	  (port-for-each (lambda (line)
			   (format out "~a\r\n"
				   (regexp-replace #/^\./ line "..")))
			 (lambda () (read-line iport)))
	  (send-command "." 250)
	  (send-command "QUIT" 221)))))))

(define-class <config> ()
  ((mailbox :init-value "~/Mail"
	    :init-keyword :mailbox)
   (mailbox-type :init-value 'mh
                 :init-keyword :mailbox-type)
   (charcode :init-value (symbol->string (gauche-character-encoding))
	     :init-keyword :charcode)
   (inbox :init-value "inbox"
	  :init-keyword :inbox)
   (smtp-host :init-value "localhost"
	      :init-keyword :smtp-host)
   (log-file :init-value "~/.scmail-log"
	     :init-keyword :log-file)
   (stdout-report :init-value #f)
   (mail-copy)
   (file-list)
   (file-name)
   ))

(define (read-config-file config-file)
  (with-error-handler
   (lambda (e) (make <config>))
   (lambda ()
     (let ((options (call-with-input-file (expand-file-name config-file)
		      (lambda (port) (read port)))))
       (apply make (cons <config> options))))))

(define-method initialize ((self <config>) initargs)
  (next-method)
  (cond ((eq? (slot-ref self 'mailbox-type) 'mh)
         (slot-set! self 'file-list mh-file-list)
         (slot-set! self 'file-name mh-file-name)
         (slot-set! self 'mail-copy mh-mail-copy))
        ((eq? (slot-ref self 'mailbox-type) 'maildir)
         (slot-set! self 'file-list maildir-file-list)
         (slot-set! self 'file-name maildir-file-name)
         (slot-set! self 'mail-copy maildir-mail-copy))
        (else
         (errorf "unsupported mailbox-type: ~a" 
                 (slot-ref self 'mailbox-type))))
  (slot-set! self 'mailbox (expand-file-name (slot-ref self 'mailbox))))

(define-method write-object ((config <config>) port)
  (format port "(config (mailbox ~a) (mailbox-type ~a) (charcode ~a) (inbox ~a) (smtp-host ~a) (log-file ~a) (stdout-report ~a))" 
	  (slot-ref config 'mailbox)
	  (slot-ref config 'mailbox-type)
	  (slot-ref config 'charcode)
	  (slot-ref config 'inbox)
	  (slot-ref config 'smtp-host)
	  (slot-ref config 'log-file)
	  (slot-ref config 'stdout-report)))

(define-class <mail> ()
  ((port :init-value (current-input-port)
	 :init-keyword :port)
   (header :init-value '())
   (charcode :init-value "ascii"
	     :init-keyword :charcode)
   (content  :init-value "")
   (info  :init-value '())
   (body  :init-value '())
   (removed?  :init-value #f)))

(define-method initialize ((self <mail>) initargs)
  (next-method)
  (mail-read self (slot-ref self 'charcode)))

(define-method write-object ((self <mail>) port)
  (format port "[mail]"))

;; (parse-header '("Subject: foo" " baz" "From: foo@bar"))
;; => ((from . "foo@bar") (subject . "foo baz"))
(define (parse-header header to-code)
  ;; (field-name "Subject: foobar") => "Subject"
  (define (field-name line)
    (substring line 0 (string-index line #\:)))

  (define (field-value line)
    (rxmatch-substring (safe-rxmatch #/\S+:\s*(.*)/ line) 1))

  ;; (concat-header '("Subject: foo" " baz" "From: foo@bar"))
  ;; => ("Subject: foo baz" "From: foo@bar") 
  (define (concat-header header)
    (define (internal old new)
      (cond ((null? old) new)
	    ((char-whitespace? (string-ref (car old) 0))
	     (set-car! new (string-append (car new) (car old)))
	     (internal (cdr old) new))
	    (else (internal (cdr old) (cons (car old) new)))))
    (internal header '()))

  (map (lambda (line)
	 (let ((name (string->symbol 
		      (string-downcase (field-name line))))
	       (value (decode-field (field-value line) to-code)))
	   (cons name value)))
       (concat-header header)))

;; Separate a header and a body from a mail.
(define (mail-separate port)
  (let ((lines (port->string-list port)))
    (if (null? lines)
        (values '() '())
        (let loop ((header '()) (lines lines))
          (cond ((null? lines)
                 (values (reverse header) '()))
                ((equal? "" (car lines))
                 (values (reverse header) (cdr lines)))
                (else (loop (cons (car lines) header) (cdr lines))))))))

(define-method mail-read ((mail <mail>) to-code)
  (define (read-mail)
    (let ((lines (port->string-list (slot-ref mail 'port))))
      (if (safe-rxmatch #/^From / (car lines)) ; Snip the first "From " line.
	  (set! lines (cdr lines)))
      (string-join lines "\n")))

  (define (add-info! key value)
    (slot-set! mail 'info
     (cons (cons key value) (slot-ref mail 'info))))

  ;; FIXME: port->string is too slow in Gauche 0.6.2
  ;;  (slot-set! mail 'content (port->string (slot-ref mail 'port)))
  (slot-set! mail 'content (read-mail))
  (receive (header body)
	   (mail-separate (open-input-string (slot-ref mail 'content)))
	   (slot-set! mail 'body body)
	   (slot-set! mail 'header header)
	   (slot-set! mail 'info (parse-header header to-code))
	   (add-info! 'length (string-length (slot-ref mail 'content)))
	   (add-info! 'to/cc
		      (string-append (mail-query mail 'to) 
				     ", " 
				     (mail-query mail 'cc)))
	   (add-info! 'body
		      (string-join body "\n"))
	   (add-info! 'file-name
		      (port-name (slot-ref mail 'port))))
  (close-input-port (slot-ref mail 'port)))

(define-method mail-query ((mail <mail>) key)
  (let ((pair (assq key (slot-ref mail 'info))))
    (if pair (cdr pair)	"")))

(define-method mail->port ((mail <mail>))
  (open-input-string (slot-ref mail 'content)))
  
(define-method mail-write ((mail <mail>) oport)
  (copy-port (mail->port mail) oport))

(define-method mail-from-stdin? ((mail <mail>))
  (equal? (slot-ref mail 'port)
	  (current-input-port)))

(define-method mail-remove ((mail <mail>))
  (slot-set! mail 'removed? #t)
  (unless (mail-from-stdin? mail)
	  (sys-unlink (mail-query mail 'file-name))))

(define-method mail-removed? ((mail <mail>))
  (slot-ref mail 'removed?))

(define (make-dest-directory mailbox folder)
  (let ((dest-directory (expand-file-name folder mailbox)))
    (create-directory* dest-directory)
    dest-directory))

(define (mail-forward mail host address)
  (send-mail host 25 (mail->port mail) "" address))

(define (scmail-log config message)
  (let ((message (if (safe-rxmatch #/\n$/ message)
                     message
                     (string-append message "\n")))
        (prefix (sys-strftime "%Y-%m-%dT%H:%M:%S: " 
			      (sys-localtime (sys-time)))))
    (if (slot-ref config 'stdout-report) (display message))
    (with-error-handler
     (lambda (e) '())
     (lambda ()
       (call-with-output-file (expand-file-name (slot-ref config 'log-file))
	 (lambda (port)
	   (display prefix port)
	   (display message port))
	 :if-exists :append)))))

(define (scmail-command-log config command src dest)
  (let ((message (format "~a: ~a -> ~a\n"
			 command src dest)))
    (scmail-log config message)))


(define (scmail-error-log config message)
  (let ((message (format "error: ~a\n" message)))
    (scmail-log config message)))


(define (cut-mailbox-part file-name mailbox)
  (let ((m (safe-rxmatch #`"^,|mailbox|/+" file-name)))
    (if m 
	(rxmatch-after m)
	file-name)))

(define (command-copy config mail folder)
  (let* ((file-name (mail-query mail 'file-name))
	 (mailbox (slot-ref config 'mailbox))
	 (src (cut-mailbox-part file-name mailbox)))
    (let* ((out-file-name (mail-copy config mail mailbox folder))
           (dest (cut-mailbox-part out-file-name mailbox)))
      (scmail-command-log config 'copy src dest)))
  :next)

(define (command-refile config mail folder)
  (let* ((file-name (mail-query mail 'file-name))
	 (mailbox (slot-ref config 'mailbox))
	 (src (cut-mailbox-part file-name mailbox)))
    (let* ((out-file-name (mail-copy config mail mailbox folder))
           (dest (cut-mailbox-part out-file-name mailbox)))
      (mail-remove mail)
      (scmail-command-log config 'refile src dest)))
  :last)

(define (command-forward config mail address)
  (let* ((file-name (mail-query mail 'file-name))
	 (mailbox (slot-ref config 'mailbox))
	 (src (cut-mailbox-part file-name mailbox)))
    (mail-forward mail (slot-ref config 'smtp-host) address)
    (mail-remove mail)
    (scmail-command-log config 'forward src address))
  :last)

(define (process-command config mail command param field-name pattern)
  (cond ((eq? command 'refile)
	 (command-refile config mail param))
	((eq? command 'copy)
	 (command-copy config mail param))
	((eq? command 'forward)
	 (command-forward config mail param))))

(define (replace-folder-name folder match)
  (regexp-replace-all #/\\([\d+])/ folder
                      (lambda (m)
                        (let ((str (rxmatch-substring 
                                    match (string->number 
                                           (rxmatch-substring m 1)))))
                          (if global-match-data-replace-rule
                              (let* ((pattern 
                                      (car global-match-data-replace-rule))
                                     (replacement 
                                      (cadr global-match-data-replace-rule)))
                                (regexp-replace-all pattern str replacement))
                              str)))))

(define (process-rule config mail field-name patterns command param)
  (if (null? patterns)
      :next
      (let* ((pattern (car patterns))
	     (status (if (regexp? pattern)
			 (let ((m (safe-rxmatch pattern 
					   (mail-query mail field-name))))
			   (if m
			       (process-command config mail command
						(replace-folder-name param m)
						field-name pattern)
			       :next))
			 (if (string-contains
			      (mail-query mail field-name) pattern)
			     (process-command config mail command param 
					      field-name pattern)
			     :next))))
	(if (eq? status :next)
	    (process-rule config mail field-name 
			  (cdr patterns) command param)
	    :last))))

(define (apply-rule config mail rule)
  (let ((field-name (car rule))
	(field-rules (cdr rule)))
    (let loop ((field-rules field-rules))
      (if (null? field-rules)
	  :next
	  (let* ((pattern-and-command (car field-rules))
		 (patterns 
		  (if (list? (first pattern-and-command))
		      (first pattern-and-command)
		      (list (first pattern-and-command))))
		 (command (if (list? (second pattern-and-command))
			      (first (second pattern-and-command))
			      'refile))
		 (param (if (list? (second pattern-and-command))
			    (second (second pattern-and-command))
			    (second pattern-and-command))))
	    (let ((status (process-rule config mail field-name
					patterns command param)))
	      (if (eq? status :next)
		  (loop (cdr field-rules))
		  :last)))))))

;; Filter a mail
(define (mail-filter config mail filter-rules)
  (unless (null? filter-rules)
	  (let* ((rule (car filter-rules))
		 (status 
		  (if (procedure? rule)
		      (rule config mail)
		      (apply-rule config mail rule))))
	    (if (eq? status :next)
		(mail-filter config mail (cdr filter-rules))))))


;; (set-match-data-replace-rule! #/\./ "-")
(define global-match-data-replace-rule #f)
(define (set-match-data-replace-rule! rule)
  (set! global-match-data-replace-rule rule))

(define global-filter-rules '())
(define (add-filter-rule! . rules)
  (set! global-filter-rules (append global-filter-rules rules)))

(define (read-filter-rule config file-name)
  (with-error-handler 
   (lambda (e)
     (scmail-error-log config (slot-ref e 'message))
     '())
   (lambda ()
     (call-with-input-file (expand-file-name file-name)
       (lambda (port) (load-from-port port)))
     global-filter-rules)))

(define (read-kill-file config file-name)
  (with-error-handler 
   (lambda (e)
     (scmail-error-log config (slot-ref e 'message))
     '())
   (lambda ()
     (call-with-input-file (expand-file-name file-name)
       (lambda (port) 
	 (let ((kill-list (port->string-list port)))
	   `((from
	      (,kill-list "junk")))))))))

(define (show-help program-name)
  (format (current-output-port) "Usage: ~a\n" program-name)
  (display "  --config=FILE    use FILE as a config file\n")
  (display "  --kill=FILE      use FILE as a kill list file\n")
  (display "  --rule=FILE      use FILE as a rule file\n")
  (display "  --folder=FOLDER  refile mails in FOLDER\n")
  (display "  --help           display this help and exit\n")
  (exit 0))

(define (scmail-main args
		     default-config-file
		     default-kill-file
		     default-rule-file
		     main-process)
		     
  (let* ((config-file default-config-file)
	 (kill-file   default-kill-file)
	 (rule-file   default-rule-file)
	 (target-folder #f))
    (parse-options (cdr args)
		   (("help" ()
                     (show-help (car args)))
                    ("config=s" (file-name)
		     (set! config-file file-name))
		    ("kill=s" (file-name)
		     (set! kill-file file-name))
		    ("rule=s" (file-name)
		     (set! rule-file file-name))
		    ("folder=s" (folder)
		     (set! target-folder folder))))
    (let* ((config (read-config-file config-file))
	   (target-folder (if target-folder
			      target-folder
			      (slot-ref config 'inbox)))
	   (target-directory 
	    (expand-file-name target-folder (slot-ref config 'mailbox)))
	   (filter-rules (append (read-filter-rule config rule-file)
				 (read-kill-file config kill-file))))
      (main-process config target-directory filter-rules))))

(provide "scmail")
